/* $Id: issub.c,v 1.2 1999/09/02 21:14:00 lindberg Exp $ */
/* $Name: ezmlm-idx-0324 $ */

/* Modified by Magnus Stlker (stalaker@umc.se) from original source
 * by Frederik Lindberg (lindberg@id.wustl.edu)
 */

#include "stralloc.h"
#include "getln.h"
#include "readwrite.h"
#include "substdio.h"
#include "open.h"
#include "byte.h"
#include "case.h"
#include "strerr.h"
#include "error.h"
#include "uint32.h"
#include "fmt.h"
#include "subscribe.h"
#include "errtxt.h"
#include <unistd.h>
#include <libpq-fe.h>

static void die_nomem(fatal)
char *fatal;
{
  strerr_die2x(111,fatal,ERR_NOMEM);
}

static stralloc addr = {0};
static stralloc lcaddr = {0};
static stralloc line = {0};
static stralloc quoted = {0};
static stralloc fn = {0};
static substdio ss;
static char ssbuf[512];

char *issub(dbname,userhost,fatal)
/* Returns (char *) to match if userhost is in the subscriber database     */
/* dbname, 0 otherwise. dbname is a base directory for a list and may NOT  */
/* be NULL        */
/* NOTE: The returned pointer is NOT VALID after a subsequent call to issub!*/

char *dbname;		/* directory to basedir */
char *userhost;
char *fatal;

{
  PGresult *result;
  char *table = "ezmlm";
  char *sublist = (char *) 0;
  char *ret;

  int fd;
  unsigned int j;
  uint32 h,lch;
  char ch,lcch;
  int match;

  if ((ret = opensql(dbname,&table,&sublist))) {
    if (*ret) strerr_die2x(111,fatal,ret);
						/* fallback to local db */

    if (!stralloc_copys(&addr,"T")) die_nomem(fatal);
    if (!stralloc_cats(&addr,userhost)) die_nomem(fatal);

    j = byte_rchr(addr.s,addr.len,'@');
    if (j == addr.len) return 0;
    case_lowerb(addr.s + j + 1,addr.len - j - 1);
    if (!stralloc_copy(&lcaddr,&addr)) die_nomem(fatal);
    case_lowerb(lcaddr.s + 1,j - 1);	/* totally lc version of addr */

    h = 5381;
    lch = h;			/* make hash for both for backwards comp */
    for (j = 0;j < addr.len;++j) {	/* (lcaddr.len == addr.len) */
      h = (h + (h << 5)) ^ (uint32) (unsigned char) addr.s[j];
      lch = (lch + (lch << 5)) ^ (uint32) (unsigned char) lcaddr.s[j];
    }
    ch = 64 + (h % 53);
    lcch = 64 + (lch % 53);

    if (!stralloc_0(&addr)) die_nomem(fatal);
    if (!stralloc_0(&lcaddr)) die_nomem(fatal);
    if (!stralloc_copys(&fn,dbname)) die_nomem(fatal);
    if (!stralloc_cats(&fn,"/subscribers/")) die_nomem(fatal);
    if (!stralloc_catb(&fn,&lcch,1)) die_nomem(fatal);
    if (!stralloc_0(&fn)) die_nomem(fatal);

    fd = open_read(fn.s);
    if (fd == -1) {
      if (errno != error_noent)
        strerr_die4sys(111,fatal,ERR_OPEN,fn.s,": ");
    } else {
      substdio_fdbuf(&ss,read,fd,ssbuf,sizeof(ssbuf));

      for (;;) {
        if (getln(&ss,&line,&match,'\0') == -1)
          strerr_die4sys(111,fatal,ERR_READ,fn.s,": ");
        if (!match) break;
        if (line.len == lcaddr.len)
          if (!case_diffb(line.s,line.len,lcaddr.s))
            { close(fd); return line.s+1; }
      }
      close(fd);
    }
	/* here if file not found or (file found && addr not there) */

    if (ch == lcch) return 0;

	/* try case sensitive hash for backwards compatibility */
    fn.s[fn.len - 2] = ch;
    fd = open_read(fn.s);
    if (fd == -1) {
      if (errno != error_noent)
        strerr_die4sys(111,fatal,ERR_OPEN,fn.s,": ");
      return 0;
    }
    substdio_fdbuf(&ss,read,fd,ssbuf,sizeof(ssbuf));

    for (;;) {
      if (getln(&ss,&line,&match,'\0') == -1)
        strerr_die4sys(111,fatal,ERR_READ,fn.s,": ");
      if (!match) break;
      if (line.len == addr.len)
        if (!case_diffb(line.s,line.len,addr.s))
          { close(fd); return line.s+1; }
    }

    close(fd);

    return 0;
  } else {						/* SQL version  */
	/* SELECT address FROM list WHERE address = 'userhost' AND hash */
	/* BETWEEN 0 AND 52. Without the hash restriction, we'd make it */
	/* even easier to defeat. Just faking sender to the list name would*/
	/* work. Since sender checks for posts are bogus anyway, I don't */
	/* know if it's worth the cost of the "WHERE ...". */

    if (!stralloc_copys(&addr,userhost)) die_nomem(fatal);
    j = byte_rchr(addr.s,addr.len,'@');
    if (j == addr.len) return 0;
    case_lowerb(addr.s + j + 1,addr.len - j - 1);
    
    if (!stralloc_0(&line)) die_nomem(fatal);
    if (!stralloc_copys(&line,"SELECT address FROM ")) die_nomem(fatal);
    if (!stralloc_cats(&line,table)) die_nomem(fatal);
    if (!stralloc_cats(&line," WHERE address ~* '^")) die_nomem(fatal);
    if (!stralloc_cats(&line, userhost)) die_nomem(fatal);
    if (!stralloc_cats(&line,"$' AND (hash BETWEEN 0 AND 52)"))
      die_nomem(fatal);
    
    /* Exec the QUERY and check for errors */
    if (!stralloc_0(&line)) die_nomem(fatal);
    /*line.s[line.len] = 0;*/
    result = PQexec( psql, line.s );
    if ( result == NULL )
      strerr_die2x(111, fatal, PQerrorMessage((PGconn *) psql));
    
    ret = (char *) 0;
    /* No data returned in QUERY */
    if( PQresultStatus( result ) != PGRES_TUPLES_OK )
      strerr_die2x( 111, fatal, PQresultErrorMessage(result));
    if (PQntuples(result) < 1)
            return 0;
    /*      strerr_die2x( 111, fatal, PQresultErrorMessage(result));*/
      
    /* Return adress in line? Maybe need some more errorchecking! */
    if (!stralloc_copys(&line, PQgetvalue(result, 0, 0)))
      die_nomem(fatal);
    /*    if (!stralloc_0(&line)) die_nomem(fatal);    */

    PQclear( result );
    ret = line.s;
    return ret;
  } /* SQL version */
}
