#include "error.h"
#include "strerr.h"
#include "readwrite.h"
#include "str.h"
#include "fmt.h"
#include "stralloc.h"
#include "open.h"
#include "substdio.h"
#include "case.h"
#include "errtxt.h"
#include "subscribe.h"
#include "qmail.h"
#include <unistd.h>
#include <libpq-fe.h>

static substdio ssin;
static char inbuf[512];
char strnum[FMT_ULONG];
static stralloc line = {0};
static stralloc domains = {0};
static stralloc quoted = {0};
static stralloc fn = {0};

static void die_nomem(fatal)
char *fatal;
{
  strerr_die2x(111,fatal,ERR_NOMEM);
}

static void die_write(fatal)
char *fatal;
{
  strerr_die3x(111,fatal,ERR_WRITE,"stdout");
}

unsigned int putsubs(dbname,userhost,flagmysql,flagusewhere,msgnum,
	subwrite,fatal)
/* Outputs all userhostesses in 'dbname' to stdout. If userhost is not null */
/* that userhost is excluded. 'dbname' is the base directory name. For the  */
/* mysql version, dbname is the directory where the file "sql" with mysql   */
/* access info is found. If this file is not present or if flagmysql is not */
/* set, the routine falls back to the old database style. subwrite must be a*/
/* function returning >=0 on success, -1 on error, and taking arguments     */
/* (char* string, unsigned int length). It will be called once per address  */
/* and should take care of newline or whatever needed for the output form.  */

char *dbname;		/* database base dir */
char *userhost;		/* sender address to be excluded */
int flagmysql;		/* if not set, ignore sql file. Needed for db */
			/* conversion capability of ezmlm-list */
int flagusewhere;	/* for mysql, use where part (for sublists) */
unsigned long msgnum;	/* for restriction */
int subwrite();		/* write function. */
char *fatal;		/* fatal error string */

{
  PGresult *result;
  char *row;
  int row_num;
  int length;

  char *table = "ezmlm";
  char *listname = (char *) 0;
  unsigned long *lengths;

  unsigned int i;
  int fd;
  unsigned int no = 0;
  int match;
  unsigned int pos = 0;
  unsigned int hashpos;
  char *ret = (char *) 0;

  if (!flagmysql || (ret = opensql(dbname,&table,&listname))) {
    if (ret && *ret) strerr_die2x(111,fatal,ret);
						/* fallback to local db */
    if (userhost && *userhost)
      pos =str_len(userhost)+2;	/* Tuserhost@host\0  used only if !flagself */
    if (!stralloc_copys(&fn,dbname)) die_nomem(fatal);
    if (!stralloc_catb(&fn,"/subscribers/?",15)) die_nomem(fatal);
				/* NOTE: Also copies terminal '\0' */
    hashpos = fn.len - 2;

    for (i = 0;i < 53;++i) {
      fn.s[hashpos] = 64 + i;
      fd = open_read(fn.s);
      if (fd == -1) {
        if (errno != error_noent)
	  strerr_die4sys(111,fatal,ERR_READ,fn.s,": ");
      } else {
        substdio_fdbuf(&ssin,read,fd,inbuf,sizeof(inbuf));
        for (;;) {
          if (getln(&ssin,&line,&match,'\0') == -1)
            strerr_die4sys(111,fatal,ERR_READ,fn.s,": ");
          if (!match)
            break;
          if ((line.len != pos) || case_diffb(userhost,pos-2,line.s+1)) {
            if (subwrite(line.s + 1,line.len - 2) == -1) die_write(fatal);
            no++;
          }
        }
        close(fd);
      }
    }
    return no;

  } else {					/* SQL Version */
/* This is a bit complicated in order to make admin and modification of  */
/* running list groups (main + sublist(s)) as easy as possible. When     */
/* there is no name, it's easy: "SELECT address FROM table [WHERE address*/
/* != sender". When there is a name and we don't use it, we do the same  */
/* thing. When there is a name and we should use it, the list either     */
/* serves a domain, or if not, should server otherwise "un-served"       */
/* domains. Since there are no subqueries for SELECT we need an extra    */
/* query to find out if domain = '' and if so, what the other domains    */
/* are. All, of course, restricted to active list (msgnum criterion      */
/* skipped if msgnum = 0). We do the latter first:                       */
/* SELECT domain FROM list_name WHERE notuse=0 AND num BETWEEN msgnum_lo */
/* AND msgnum_hi. Instead of adding "AND domain != ''" we just eliminate */
/* that row ourselves. From this info, we build a string: 'd1','d2',...  */
/* Then:                                                                 */
/* SELECT address FROM list AS t1,list_name AS t2 WHERE [address !=      */
/* sender AND] notuse=0 AND num BETWEEN msgnum_lo AND msgnum_hi AND      */
/* (t1.domain = t2.domain OR t2.domain = '' AND NOT t1.domain in (the    */
/* string of explicitly served domains)) AND hash BETWEEN hash_lo AND    */
/* hash_hi. The list_name table should have few records (<10 for most    */
/* lists, never >100 even for sloppily admin'ed large lists). HOWEVER,   */
/* we need to allow for multiple entries for a give list name to e.g. use*/
/* one list to serve 'se', 'no', and 'fi'. This is the query worth       */
/* maximally optimizing and for SQL servers with subselects, we could get*/
/* away with a single query. */
/* After this, if we used the restrictions, we add addresses with hash=98*/
/* which are used for feedback to verify that the sublist deliveres mail.*/

    flagusewhere = (flagusewhere && listname);
    if (!stralloc_copys(&domains,"")) die_nomem(fatal);		/* clear */
    if (flagusewhere) {					/* list of domains */
	  if (!stralloc_0(&line)) die_nomem(fatal);
	  if (!stralloc_copys(&line,"SELECT DISTINCT domain FROM "))
		die_nomem(fatal);
      if (!stralloc_cats(&line,table)) die_nomem(fatal);
      if (!stralloc_cats(&line,"_name WHERE notuse=0")) die_nomem(fatal);
      if (msgnum) {
        if (!stralloc_cats(&line," AND ")) die_nomem(fatal);
        if (!stralloc_cats(&line,strnum,fmt_ulong(strnum,msgnum)))
		  die_nomem(fatal);
        if (!stralloc_cats(&line," BETWEEN msgnum_lo AND msgnum_hi"))
		  die_nomem(fatal);
      }

      if (!stralloc_ready(&domains,125)) die_nomem(fatal);	/* 20x6+1+chg*/
      if (!stralloc_copys(&domains,"'")) die_nomem(fatal);
      no = 0;
	  line.s[line.len] = 0;
	  result = PQexec( psql, line.s );
	  if ( result == NULL )
	    strerr_die2x( 111, fatal, PQerrorMessage((PGconn *) psql));
	  if ( PQresultStatus(result) != PGRES_TUPLES_OK )
	    strerr_die2x( 111, fatal, PQresultErrorMessage(result));

	  if ( PQresultStatus(result) == PGRES_TUPLES_OK )
	    for ( row_num = 0; row_num < PQntuples(result); row_num++ ) {
	        row = PQgetvalue( result, row_num, 0 );
		if (row[1] == '\0') continue;
		/* zap domain = '' */
		/* possible addresses with domain ='' will be caught by the*/
		/* test for equivalence in the query */

		if (!stralloc_cats(&domains,row[0]))
		  die_nomem(fatal);
		if (!stralloc_cats(&domains,"','")) die_nomem(fatal);
		if (++no % 20 == 0)
		  if (!stralloc_readyplus(&domains,120)) die_nomem(fatal);
	    }
	  PQclear( result );
	  if (domains.len >= 2) domains.len -= 2;	/* remove last ,' */
    }
						/* main query */
    if (!stralloc_0(&line)) die_nomem(fatal);
    if (!stralloc_copys(&line,"SELECT address FROM "))
	  die_nomem(fatal);
    if (!stralloc_cats(&line,table)) die_nomem(fatal);
    if (flagusewhere) {		/* name table only is we use the name */
      if (!stralloc_cats(&line," AS t1,")) die_nomem(fatal);
      if (!stralloc_cats(&line,table)) die_nomem(fatal);
      if (!stralloc_cats(&line,"_name AS t2")) die_nomem(fatal);
    }
    if (userhost && *userhost) {
      if (!stralloc_cats(&line," WHERE address !~* '^")) die_nomem(fatal);
      i = str_len(userhost);
      if (!stralloc_catb(&line, userhost, i)) die_nomem(fatal);
      if (!stralloc_cats(&line,"$'")) die_nomem(fatal);
    }
    if (flagusewhere) {
	if (userhost && *userhost) {	/* eliminate SENDER if requested */
	  if (!stralloc_cats(&line," AND ")) die_nomem(fatal);
	} else {
	  if (!stralloc_cats(&line," WHERE ")) die_nomem(fatal);
	}
	if (!stralloc_cats(&line,"name ~* '^")) die_nomem(fatal);
	if (!stralloc_cats(&line,listname)) die_nomem(fatal);
	if (!stralloc_cats(&line,
		"$' AND notuse = 0 AND (hash BETWEEN hash_lo AND hash_hi) "))
	  die_nomem(fatal);
	if (domains.len > 1) {	/* no domains => no domain restrictions */
	  if (!stralloc_cats(&line,"AND (t1.domain = t2.domain OR ("
			   "t2.domain = '' AND NOT (t1.domain IN ("))
	    die_nomem(fatal);
	  if (!stralloc_cat(&line,&domains)) die_nomem(fatal);
	  if (!stralloc_cats(&line,"))))")) die_nomem(fatal);
        }
	if (msgnum) {		/* msgnum=0 => no msgnum restrictions */
	  if (!stralloc_cats(&line," AND ")) die_nomem(fatal);
	  if (!stralloc_cats(&line,strnum,fmt_ulong(strnum,msgnum)))
	    die_nomem(fatal);
	  if (!stralloc_cats(&line," BETWEEN msgnum_lo AND msgnum_hi"))
	    die_nomem(fatal);
	}
    }
    if (!stralloc_0(&line)) die_nomem(fatal);
    /*line.s[line.len] = 0;*/
    result = PQexec( psql, line.s );
    if ( result == NULL )
      strerr_die2x( 111, fatal, PQerrorMessage((PGconn *) psql));
    if ( PQresultStatus(result) != PGRES_TUPLES_OK )
      strerr_die2x( 111, fatal, PQresultErrorMessage(result));

    no = 0;
    for ( row_num = 0; row_num < PQntuples(result); row_num++ ) {
	row = PQgetvalue(result,row_num,0);
	/* this is safe even if someone messes with the address field def */
	length = PQgetlength(result,row_num,0);
	if (subwrite(row, length) == -1) die_write(fatal);
	no++;					/* count for list-list fxn */
    }
    PQclear(result);

	/* and now the feedback addresses (sep query - otherwise we'd need */
	/* ORDER BY hash, which for a list of any size would be much more  */
	/* expensive. If !flagusewhere, we've got them  already. */

    if (flagusewhere) {
      if (!stralloc_0(&line)) die_nomem(fatal);
      if (!stralloc_copys(&line,"SELECT address FROM "))
		die_nomem(fatal);
      if (!stralloc_cats(&line,table)) die_nomem(fatal);
      if (!stralloc_cats(&line," WHERE hash=98")) die_nomem(fatal);
      if (!stralloc_0(&line)) die_nomem(fatal);
      /*      line.s[line.len] = 0; */
      result = PQexec( psql, line.s );
      if( result == NULL )
	strerr_die2x( 111, fatal, PQerrorMessage((PGconn *) psql));
      if ( PQresultStatus(result) != PGRES_TUPLES_OK )
	return 0;

      for ( row_num = 0; row_num < PQntuples( result ); row_num++ ) {
	row = PQgetvalue(result,row_num,0);
	length = PQgetlength(result,row_num,0);
	if (subwrite(row, length) == -1) die_write(fatal);
	no++;						/* count */
      }
      PQclear( result );
    }
    return no;
  }
}
